#!/usr/bin/env python

import argparse
import glob
import numpy as np
import os
import time
import cv2 as cv
import torch
from utils.reconstruct import TPS_line
from utils.utils import load_model
from modeling.TP_Net import Res160, Res320
from modeling.Hourglass import HourglassNet
import sys
import cairosvg
import iio


def normalized_read(img_path, resize):
    img = cv.imread(img_path)
    if resize:
        inp = cv.resize(img, (512, 512))
    else:
        inp = img

    H, W, C = inp.shape
    hsv = cv.cvtColor(inp, cv.COLOR_BGR2HSV)
    imgv0 = hsv[..., 2]
    imgv = cv.resize(imgv0, (0, 0), fx=1. / 4, fy=1. / 4, interpolation=cv.INTER_LINEAR)
    imgv = cv.GaussianBlur(imgv, (5, 5), 3)
    imgv = cv.resize(imgv, (W, H), interpolation=cv.INTER_LINEAR)
    imgv = cv.GaussianBlur(imgv, (5, 5), 3)

    imgv1 = imgv0.astype(np.float32) - imgv + 127.5
    imgv1 = np.clip(imgv1, 0, 255).astype(np.uint8)
    hsv[..., 2] = imgv1
    inp = cv.cvtColor(hsv, cv.COLOR_HSV2BGR)

    inp = (inp.astype(np.float32) / 255.)
    inp = inp.transpose(2, 0, 1)
    inp = torch.from_numpy(inp).unsqueeze(0)
    return inp

if __name__ == '__main__':
    
    img_path = sys.argv[1]
    out = sys.argv[2]
    model_ = sys.argv[3]
    thresh = float(sys.argv[4])

    out_svg = out + '.svg'
    out_txt = out + '.txt'
    out_png = out + '.png'

    # Load model
    head = {'center': 1, 'dis': 4, 'line': 1}
    assert model_ in ['tp320', 'tplite', 'tp512', 'hg'], "Incompatible model"
    if model_ == 'tp320':
        model = load_model(Res320(head), '/workdir/demoextras/pretraineds/Res320.pth')
    elif model_ == 'tplite':
        model = load_model(Res160(head), '/workdir/demoextras/pretraineds/Res160.pth')
    elif model_ == 'tp512':
        model = load_model(Res320(head), '/workdir/demoextras/pretraineds/Res512.pth')
    elif model_ == 'hg':
        model = load_model(HourglassNet(head), '/workdir/demoextras/pretraineds/HG128.pth')

    imgn = normalized_read(img_path, resize=(model_=='hg'))
    B, C, H, W = imgn.size()

    img = iio.read(img_path)
    h, w, c = img.shape

    with torch.no_grad():
      outputs = model(imgn)

    output = outputs[-1]
    # Add a parameter for lmbd?
    lines, scores, start_point, end_point, pos, endtime = TPS_line(output, thresh, 0.5, H, W)

    if model_ == 'hg':
        print(w,W,h,H)
        sx, sy = w / W, h / H
        print(sx,sy)
        print(lines)
        lines[:, [0, 2]] *= sx
        lines[:, [1, 3]] *= sy

    ##### OUTPUT ######

    ftxt = open(out_txt, "w")
    fcsv = open(out_svg, "w")

    fcsv.write('<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"\n"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"><svg version="1.1" xmlns="http://www.w3.org/2000/svg"\nxmlns:xlink="http://www.w3.org/1999/xlink"\nwidth="'+str(w)+'" height="'+str(h)+'" viewBox="0 0 '+str(w)+' '+str(h)+'">')
    topng = '<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"\n"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"><svg version="1.1" xmlns="http://www.w3.org/2000/svg"\nxmlns:xlink="http://www.w3.org/1999/xlink"\nwidth="'+str(w)+'" height="'+str(h)+'" viewBox="0 0 '+str(w)+' '+str(h)+'">'

    color='black'
    width=1.5
    for tp_id, (line, score) in enumerate(zip(lines, scores)):
        x1, y1, x2, y2 = line # this is yxyx
        fcsv.write('<line x1="'+str(x1.item())+'" y1="' +str(y1.item())+ '" x2="' +str(x2.item())+ '" y2="' +str(y2.item())+ '" stroke="'+color+'" stroke-width="'+str(width)+'"/>\n')
        topng += '<line x1="'+str(x1.item())+'" y1="' +str(y1.item())+ '" x2="' +str(x2.item())+ '" y2="' +str(y2.item())+ '" stroke="white" stroke-width="'+str(width)+'"/>\n'
        ftxt.write(str(x1.item())+' '+str(y1.item())+' '+str(x2.item())+' '+str(y2.item())+' '+str(score)+'\n')
    fcsv.write('\n</svg>')
    topng += '\n</svg>'
    ftxt.close() 
    fcsv.close() 

    cairosvg.svg2png(bytestring=topng, write_to=out_png)
    im = iio.read(out_png)
    iio.write(out_png, im[:,:,:3])

